@@ -17,9 +17,11 @@ setClass("RangedSummarizedExperiment",
     )
 )
 
-### Combine the new parallel slots with those of the parent class. Make sure
-### to put the new parallel slots *first*.
-setMethod("parallelSlotNames", "RangedSummarizedExperiment",
+### Combine the new "vertical slots" with those of the parent class. Make
+### sure to put the new vertical slots **first**. See R/bindROWS.R file in
+### the S4Vectors package for what slots should or should not be considered
+### "vertical".
+setMethod("vertical_slot_names", "RangedSummarizedExperiment",
     function(x) c("rowRanges", callNextMethod())
 )
 
@@ -58,9 +60,8 @@ setValidity2("RangedSummarizedExperiment", .valid.RangedSummarizedExperiment)
 .new_RangedSummarizedExperiment <- function(assays, rowRanges, colData,
                                             metadata)
 {
+    assays <- Assays(assays, as.null.if.zero.assay=TRUE)
     elementMetadata <- S4Vectors:::make_zero_col_DataFrame(length(rowRanges))
-    if (!is(assays, "Assays"))
-        assays <- Assays(assays)
     new("RangedSummarizedExperiment", rowRanges=rowRanges,
                                       colData=colData,
                                       assays=assays,
@@ -68,13 +69,6 @@ setValidity2("RangedSummarizedExperiment", .valid.RangedSummarizedExperiment)
                                       metadata=as.list(metadata))
 }
 
-.get_colnames_from_assays <- function(assays)
-{
-    if (length(assays) == 0L)
-        return(NULL)
-    colnames(assays[[1L]])
-}
-
 .get_rownames_from_assays <- function(assays)
 {
     if (length(assays) == 0L)
@@ -82,21 +76,29 @@ setValidity2("RangedSummarizedExperiment", .valid.RangedSummarizedExperiment)
     rownames(assays[[1L]])
 }
 
+.get_colnames_from_assays <- function(assays)
+{
+    if (length(assays) == 0L)
+        return(NULL)
+    colnames(assays[[1L]])
+}
+
 SummarizedExperiment <- function(assays=SimpleList(),
                                  rowData=NULL, rowRanges=GRangesList(),
                                  colData=DataFrame(),
                                  metadata=list())
 {
-    assays <- normarg_assays(assays)
-    if (missing(colData) && 0L != length(assays)) {
-        assay <- assays[[1]]
-        nms <- colnames(assay)
-        colData <- DataFrame(x=seq_len(ncol(assay)), row.names=nms)[, FALSE]
-    } else if (!missing(colData)) {
+    assays <- normarg_assays(assays, as.null.if.zero.assay=TRUE)
+
+    if (!missing(colData)) {
         if (!is(colData, "DataFrame"))
             colData <- as(colData, "DataFrame")
         if (is.null(rownames(colData)))
             rownames(colData) <- .get_colnames_from_assays(assays)
+    } else if (length(assays) != 0L) {
+        a1 <- assays[[1L]]
+        nms <- colnames(a1)
+        colData <- DataFrame(x=seq_len(ncol(a1)), row.names=nms)[, FALSE]
     }
 
     if (is.null(rowData)) {
@@ -124,8 +126,6 @@ SummarizedExperiment <- function(assays=SimpleList(),
         }
     }
 
-    assays <- Assays(assays)
-
     if (missing(rowRanges) && !is(rowData, "GenomicRanges_OR_GRangesList")) {
         new_SummarizedExperiment(assays, ans_rownames, rowData, colData,
                                  metadata)
@@ -201,8 +201,11 @@ setReplaceMethod("rowRanges", c("RangedSummarizedExperiment", "NULL"),
 .SummarizedExperiment.rowRanges.replace <-
     function(x, ..., value)
 {
-    if (!is(x, "RangedSummarizedExperiment"))
+    if (is(x, "RangedSummarizedExperiment")) {
+        x <- updateObject(x, check=FALSE)
+    } else {
         x <- as(x, "RangedSummarizedExperiment")
+    }
     x <- BiocGenerics:::replaceSlots(x, ...,
              rowRanges=value,
              elementMetadata=S4Vectors:::make_zero_col_DataFrame(length(value)),
@@ -234,7 +237,8 @@ setReplaceMethod("names", "RangedSummarizedExperiment",
 setMethod("dimnames", "RangedSummarizedExperiment",
     function(x)
 {
-    list(names(x), rownames(colData(x)))
+    ans <- list(names(x), rownames(colData(x)))
+    DelayedArray:::simplify_NULL_dimnames(ans)
 })
 
 setReplaceMethod("dimnames", c("RangedSummarizedExperiment", "list"),
