@@ -4,23 +4,24 @@
 ###
 
 
+setClassUnion("Assays_OR_NULL", c("Assays", "NULL"))
+
 setClass("SummarizedExperiment",
     contains="Vector",
     representation(
-        colData="DataFrame",              # columns and their annotations
-        assays="Assays",                  # Data -- e.g., list of matricies
+        colData="DataFrame",            # columns and their annotations
+        assays="Assays_OR_NULL",        # Data -- e.g., list of matrices
         NAMES="character_OR_NULL",
         elementMetadata="DataFrame"
-    ),
-    prototype(
-        assays=Assays()
     )
 )
 
-### Combine the new parallel slots with those of the parent class. Make sure
-### to put the new parallel slots *first*.
-setMethod("parallelSlotNames", "SummarizedExperiment",
-    function(x) c("NAMES", callNextMethod())
+### Combine the new "vertical slots" with those of the parent class. Make
+### sure to put the new vertical slots **first**. See R/bindROWS.R file in
+### the S4Vectors package for what slots should or should not be considered
+### "vertical".
+setMethod("vertical_slot_names", "SummarizedExperiment",
+    function(x) c("assays", "NAMES", callNextMethod())
 )
 
 
@@ -79,13 +80,15 @@ setValidity2("SummarizedExperiment", .valid.SummarizedExperiment)
 new_SummarizedExperiment <- function(assays, names, rowData, colData,
                                      metadata)
 {
-    if (!is(assays, "Assays"))
-        assays <- Assays(assays)
+    assays <- Assays(assays, as.null.if.zero.assay=TRUE)
     if (is.null(rowData)) {
-        if (is.null(names))
-            nrow <- nrow(assays)
-        else
+        if (!is.null(names)) {
             nrow <- length(names)
+        } else if (!is.null(assays)) {
+            nrow <- nrow(assays)
+        } else {
+            nrow <- 0L
+        }
         rowData <- S4Vectors:::make_zero_col_DataFrame(nrow)
     } else {
         rownames(rowData) <- NULL
@@ -124,6 +127,7 @@ setGeneric("rowData", signature="x",
     function(x, use.names=TRUE, ...) standardGeneric("rowData")
 )
 
+### Fix old DataFrame instances on-the-fly (mcols() does it).
 setMethod("rowData", "SummarizedExperiment",
     function(x, use.names=TRUE, ...) mcols(x, use.names=use.names, ...)
 )
@@ -137,7 +141,10 @@ setReplaceMethod("rowData", "SummarizedExperiment",
 
 setGeneric("colData", function(x, ...) standardGeneric("colData"))
 
-setMethod("colData", "SummarizedExperiment", function(x, ...) x@colData)
+### Fix old DataFrame instances on-the-fly.
+setMethod("colData", "SummarizedExperiment",
+    function(x, ...) updateObject(x@colData, check=FALSE)
+)
 
 setGeneric("colData<-",
     function(x, ..., value) standardGeneric("colData<-"))
@@ -147,6 +154,7 @@ setReplaceMethod("colData", c("SummarizedExperiment", "DataFrame"),
 {
     if (nrow(value) != ncol(x))
         stop("nrow of supplied 'colData' must equal ncol of object")
+    x <- updateObject(x, check=FALSE)
     BiocGenerics:::replaceSlots(x, colData=value, check=FALSE)
 })
 
@@ -154,37 +162,90 @@ setReplaceMethod("colData", c("SummarizedExperiment", "NULL"),
     function(x, ..., value)
 {
     value <- new2("DFrame", nrows=ncol(x), rownames=colnames(x), check=FALSE)
+    x <- updateObject(x, check=FALSE)
     BiocGenerics:::replaceSlots(x, colData=value, check=FALSE)
 })
 
-setGeneric("assays",
-    function(x, ..., withDimnames=TRUE) standardGeneric("assays"),
-    signature="x")
+setGeneric("assays", signature="x",
+    function(x, withDimnames=TRUE, ...) standardGeneric("assays")
+)
 
 setMethod("assays", "SummarizedExperiment",
-    function(x, ..., withDimnames=TRUE)
+    function(x, withDimnames=TRUE, ...)
 {
+    if (!isTRUEorFALSE(withDimnames))
+        stop(wmsg("'withDimnames' must be TRUE or FALSE"))
     assays <- as(x@assays, "SimpleList")
     if (withDimnames) {
+        x_dimnames <- dimnames(x)
+        if (is.null(x_dimnames))
+            x_dimnames <- list(NULL, NULL)
         assays <- endoapply(assays,
-            function(assay) {
-                dimnames(assay)[1:2] <- dimnames(x)
-                assay
+            function(a) {
+                a_dimnames <- dimnames(a)
+                a_dimnames[1:2] <- x_dimnames
+                dimnames(a) <- DelayedArray:::simplify_NULL_dimnames(a_dimnames)
+                a
             }
         )
     }
     assays
 })
 
-setGeneric("assays<-",
-    function(x, ..., withDimnames=TRUE, value) standardGeneric("assays<-"),
-    signature=c("x", "value"))
+setGeneric("assays<-", signature=c("x", "value"),
+    function(x, withDimnames=TRUE, ..., value) standardGeneric("assays<-"),
+)
+
+### For assays with more than 2 dimensions, **only** the first 2 dimnames
+### components (i.e. rownames and colnames) are compared with 'x_dimnames'.
+.assays_have_identical_dimnames <- function(assays, x_dimnames)
+{
+    if (length(assays) == 0L)
+        return(TRUE)
+    ## Like DelayedArray:::simplify_NULL_dimnames() but first replaces list
+    ## elements that are equal to character(0) with NULLs.
+    normalize_dimnames <- function(dn) {
+        empty_idx <- which(lengths(dn) == 0L)
+        if (length(empty_idx) == length(dn))
+            return(NULL)
+        dn[empty_idx] <- list(NULL)
+        dn
+    }
+    x_dimnames <- normalize_dimnames(x_dimnames)
+    ok <- vapply(seq_along(assays),
+        function(i) {
+            a <- getListElement(assays, i)
+            a_dimnames <- normalize_dimnames(dimnames(a))
+            identical(a_dimnames, x_dimnames)
+        },
+        logical(1)
+    )
+    all(ok)
+}
 
 .SummarizedExperiment.assays.replace <-
-    function(x, ..., withDimnames=TRUE, value)
+    function(x, withDimnames=TRUE, ..., value)
 {
-    ## withDimnames arg allows names(assays(se, withDimnames=FALSE)) <- value
-    x <- BiocGenerics:::replaceSlots(x, assays=Assays(value), check=FALSE)
+    if (!isTRUEorFALSE(withDimnames))
+        stop(wmsg("'withDimnames' must be TRUE or FALSE"))
+    value <- normarg_assays(value, as.null.if.zero.assay=TRUE)
+    ## By default the dimnames on the supplied assays must be identical to
+    ## the dimnames on 'x'. The user must use 'withDimnames=FALSE' if it's
+    ## not the case. This is for symetry with the behavior of the getter.
+    ## See https://github.com/Bioconductor/SummarizedExperiment/issues/35
+    if (withDimnames && !.assays_have_identical_dimnames(value, dimnames(x))) {
+        if (length(dim(value)) > 2L) {
+            what <- "rownames and colnames"
+        } else {
+            what <- "dimnames"
+        }
+        stop(wmsg("please use 'assay(x, withDimnames=FALSE)) <- value' ",
+                  "or 'assays(x, withDimnames=FALSE)) <- value' when the ",
+                  what, " on the supplied assay(s) are not identical to ",
+                  "the ", what, " on ", class(x), " object 'x'"))
+    }
+    new_assays <- Assays(value, as.null.if.zero.assay=TRUE)
+    x <- BiocGenerics:::replaceSlots(x, assays=new_assays, check=FALSE)
     ## validObject(x) should NOT be called below because it would then
     ## fully re-validate objects that derive from SummarizedExperiment
     ## (e.g. DESeqDataSet objects) after the user sets the assays slot with
@@ -206,13 +267,15 @@ setReplaceMethod("assays", c("SummarizedExperiment", "SimpleList"),
 setReplaceMethod("assays", c("SummarizedExperiment", "list"),
     .SummarizedExperiment.assays.replace)
 
-setGeneric("assay", function(x, i, ...) standardGeneric("assay"))
+setGeneric("assay", signature=c("x", "i"),
+    function(x, i, withDimnames=TRUE, ...) standardGeneric("assay")
+)
 
 ## convenience for common use case
 setMethod("assay", c("SummarizedExperiment", "missing"),
-    function(x, i, ...)
+    function(x, i, withDimnames=TRUE, ...)
 {
-    assays <- assays(x, ...)
+    assays <- assays(x, withDimnames=withDimnames, ...)
     if (0L == length(assays))
         stop("'assay(<", class(x), ">, i=\"missing\", ...) ",
              "length(assays(<", class(x), ">)) is 0'")
@@ -220,10 +283,10 @@ setMethod("assay", c("SummarizedExperiment", "missing"),
 })
 
 setMethod("assay", c("SummarizedExperiment", "numeric"),
-    function(x, i, ...)
+    function(x, i, withDimnames=TRUE, ...)
 {
     tryCatch({
-        assays(x, ...)[[i]]
+        assays(x, withDimnames=withDimnames, ...)[[i]]
     }, error=function(err) {
         stop("'assay(<", class(x), ">, i=\"numeric\", ...)' ",
              "invalid subscript 'i'\n", conditionMessage(err))
@@ -231,12 +294,12 @@ setMethod("assay", c("SummarizedExperiment", "numeric"),
 })
 
 setMethod("assay", c("SummarizedExperiment", "character"),
-    function(x, i, ...)
+    function(x, i, withDimnames=TRUE, ...)
 {
     msg <- paste0("'assay(<", class(x), ">, i=\"character\", ...)' ",
                   "invalid subscript 'i'")
     res <- tryCatch({
-        assays(x, ...)[[i]]
+        assays(x, withDimnames=withDimnames, ...)[[i]]
     }, error=function(err) {
         stop(msg, "\n", conditionMessage(err))
     })
@@ -246,29 +309,29 @@ setMethod("assay", c("SummarizedExperiment", "character"),
 })
 
 setGeneric("assay<-", signature=c("x", "i"),
-    function(x, i, ..., value) standardGeneric("assay<-"))
+    function(x, i, withDimnames=TRUE, ..., value) standardGeneric("assay<-"))
 
 setReplaceMethod("assay", c("SummarizedExperiment", "missing"),
-    function(x, i, ..., value)
+    function(x, i, withDimnames=TRUE, ..., value)
 {
-    if (0L == length(assays(x)))
+    if (0L == length(assays(x, withDimnames=FALSE)))
         stop("'assay(<", class(x), ">) <- value' ", "length(assays(<",
              class(x), ">)) is 0")
-    assays(x)[[1]] <- value
+    assays(x, withDimnames=withDimnames, ...)[[1]] <- value
     x
 })
 
 setReplaceMethod("assay", c("SummarizedExperiment", "numeric"),
-    function(x, i = 1, ..., value)
+    function(x, i, withDimnames=TRUE, ..., value)
 {
-    assays(x, ...)[[i]] <- value
+    assays(x, withDimnames=withDimnames, ...)[[i]] <- value
     x
 })
 
 setReplaceMethod("assay", c("SummarizedExperiment", "character"),
-    function(x, i, ..., value)
+    function(x, i, withDimnames=TRUE, ..., value)
 {
-    assays(x, ...)[[i]] <- value
+    assays(x, withDimnames=withDimnames, ...)[[i]] <- value
     x
 })
 
@@ -300,7 +363,8 @@ setMethod("dim", "SummarizedExperiment",
 setMethod("dimnames", "SummarizedExperiment",
     function(x)
 {
-    list(names(x), rownames(colData(x)))
+    ans <- list(names(x), rownames(colData(x)))
+    DelayedArray:::simplify_NULL_dimnames(ans)
 })
 
 setReplaceMethod("dimnames", c("SummarizedExperiment", "list"),
@@ -336,6 +400,7 @@ setReplaceMethod("dimnames", c("SummarizedExperiment", "NULL"),
     idx
 }
 
+### TODO: Refactor this to use extractROWS().
 setMethod("[", c("SummarizedExperiment", "ANY", "ANY"),
     function(x, i, j, ..., drop=TRUE)
 {
@@ -409,6 +474,7 @@ setMethod("[", c("SummarizedExperiment", "ANY", "ANY"),
     ans
 })
 
+### TODO: Refactor this to use replaceROWS().
 setReplaceMethod("[",
     c("SummarizedExperiment", "ANY", "ANY", "SummarizedExperiment"),
     function(x, i, j, ..., value)
@@ -523,23 +589,6 @@ setReplaceMethod("[",
     ans
 })
 
-setMethod("extractROWS", "SummarizedExperiment",
-    function(x, i)
-    {
-        i <- normalizeSingleBracketSubscript(i, x)
-        x[i, ]
-    }
-)
-
-setMethod("replaceROWS", "SummarizedExperiment",
-    function(x, i, value)
-    {
-        i <- normalizeSingleBracketSubscript(i, x)
-        x[i, ] <- value
-        x
-    }
-)
-
 setMethod("subset", "SummarizedExperiment",
     function(x, subset, select, ...)
 {
@@ -606,16 +655,16 @@ setMethod("show", "SummarizedExperiment",
     coolcat("assays(%d): %s\n", nms)
 
     ## rownames()
-    dimnames <- dimnames(object)
-    dlen <- sapply(dimnames, length)
-    if (dlen[[1]]) coolcat("rownames(%d): %s\n", dimnames[[1]])
+    rownames <- rownames(object)
+    if (!is.null(rownames)) coolcat("rownames(%d): %s\n", rownames)
     else cat("rownames: NULL\n")
 
     ## rowData`()
     coolcat("rowData names(%d): %s\n", names(rowData(object, use.names=FALSE)))
 
     ## colnames()
-    if (dlen[[2]]) coolcat("colnames(%d): %s\n", dimnames[[2]])
+    colnames <- colnames(object)
+    if (!is.null(colnames)) coolcat("colnames(%d): %s\n", colnames)
     else cat("colnames: NULL\n")
 
     ## colData()
